@@ -30,6 +30,7 @@
 
 #include "file-data.h"
 #include "file-utils.h"
+#include "glib-utils.h"
 #include "fr-command.h"
 #include "fr-command-arj.h"
 
@@ -141,7 +142,7 @@ list__process_line (char     *line,
 		/* read file info. */
 		
 		fields = split_line (line, 7);
-		fdata->size = atoll (fields[2]);
+		fdata->size = g_ascii_strtoull (fields[2], NULL, 10);
 		fdata->modified = mktime_from_string (fields[5], fields[6]); 
 		g_strfreev (fields);
 		
@@ -280,7 +281,8 @@ fr_command_arj_extract (FRCommand  *comm,
 		char *swtch = g_strconcat ("-g/", password, NULL);
 		fr_process_add_arg (comm->process, swtch);
 		g_free (swtch);
-	}
+	} else
+ 		fr_process_add_arg (comm->process, "-g/");
 
 	fr_process_add_arg (comm->process, "-i");
 	fr_process_add_arg (comm->process, "-y");
@@ -318,7 +320,12 @@ static void
 fr_command_arj_handle_error (FRCommand   *comm, 
 			     FRProcError *error)
 {
-	/* FIXME */
+	if (error->type == FR_PROC_ERROR_GENERIC) {
+ 		if (error->status <= 1)
+ 			error->type = FR_PROC_ERROR_NONE;
+		else if (error->status == 3) 
+ 			error->type = FR_PROC_ERROR_ASK_PASSWORD;
+ 	}
 }
 
 
@@ -345,6 +352,8 @@ fr_command_arj_class_init (FRCommandArjClass *class)
 static void 
 fr_command_arj_init (FRCommand *comm)
 {
+	comm->file_type = FR_FILE_TYPE_ARJ;
+
 	comm->propCanModify                = TRUE;
 	comm->propAddCanUpdate             = TRUE; 
 	comm->propAddCanReplace            = TRUE; 
