@@ -30,6 +30,7 @@
 
 #include "file-data.h"
 #include "file-utils.h"
+#include "glib-utils.h"
 #include "fr-command.h"
 #include "fr-command-zoo.h"
 
@@ -344,6 +345,8 @@ fr_command_zoo_class_init (FRCommandZooClass *class)
 static void 
 fr_command_zoo_init (FRCommand *comm)
 {
+	comm->file_type = FR_FILE_TYPE_ZOO;
+
 	comm->propAddCanUpdate             = TRUE;
 	comm->propAddCanReplace            = FALSE; 
 	comm->propExtractCanAvoidOverwrite = FALSE;
